/*
Transforme de Fourier

Copyright (C) 2004 Observatoire de Paris-Meudon

Ce programme est un logiciel libre ; vous pouvez le redistribuer et/ou le modifier conformment aux dispositions de la Licence Publique Gnrale GNU, telle que publie par la Free Software Foundation ; version 2 de la licence, ou encore ( votre choix) toute version ultrieure.

Ce programme est distribu dans l'espoir qu'il sera utile, mais SANS AUCUNE GARANTIE ; sans mme la garantie implicite de COMMERCIALISATION ou D'ADAPTATION A UN OBJET PARTICULIER. Pour plus de dtail, voir la Licence Publique Gnrale GNU.

Vous devez avoir reu un exemplaire de la Licence Publique Gnrale GNU en mme temps que ce programme ; si ce n'est pas le cas, crivez  la Free Software Foundation Inc., 675 Mass Ave, Cambridge, MA 02139, Etats-Unis.
*/

import java.io.*;
import java.net.*;

import simu.*;
import TFLib;

/**
 * Algorithme de simulation pour SimuApplet.
 * Transforme de Fourier
 * compilation: javac13 -target 1.1 -encoding ISO-8859-1 -d build/TF -classpath SimuLab.jar TF.java
 */
public class TF extends SimuApplet {
    
    Affichage graphe1;
    Affichage graphe2;
    String dossier;
    
    public void init() {
        ParamIn fonction = new ParamIn();
        fonction.type = "string";
        fonction.label = "fonction";
        fonction.titre = "fonction";
        fonction.acquisition = "choix";
        String[] fchoix = {"porte", "crneau", "sinusode", "2 sinusodes"};
        fonction.choix = fchoix;
        fonction.defaut = "porte";
        addParamIn(fonction);
        
        ParamIn duree = new ParamIn();
        duree.type = "nombre";
        duree.label = "duree";
        duree.titre = "dure";
        duree.unit = "s";
        duree.acquisition = "champ";
        duree.defaut = "2";
        addParamIn(duree);
		
        ParamIn periode = new ParamIn();
        periode.type = "nombre";
        periode.label = "periode";
        periode.titre = "priode";
        periode.unit = "s";
        periode.acquisition = "champ";
        periode.defaut = "1";
        addParamIn(periode);
		
        ParamIn periode2 = new ParamIn();
        periode2.type = "nombre";
        periode2.label = "periode2";
        periode2.titre = "priode 2";
        periode2.unit = "s";
        periode2.acquisition = "champ";
        periode2.defaut = "0.2";
        addParamIn(periode2);
		
        ParamIn npts = new ParamIn();
        npts.type = "nombre";
        npts.label = "npts";
        npts.titre = "chantillon";
        npts.unit = "points";
        npts.acquisition = "champ";
        npts.defaut = "128";
        addParamIn(npts);
                
        ParamIn methode = new ParamIn();
        methode.type = "string";
        methode.label = "methode";
        methode.titre = "mthode";
        methode.acquisition = "choix";
        String[] mchoix = {"FFT", "FT"};
        methode.choix = mchoix;
        methode.defaut = "FFT";
        addParamIn(methode);
        
        ParamIn amplitude = new ParamIn();
        amplitude.type = "nombre";
        amplitude.label = "amplitude";
        amplitude.titre = "amplitude";
        amplitude.acquisition = "champ";
        amplitude.defaut = "1";
        addParamIn(amplitude);
                
        ParamIn phase = new ParamIn();
        phase.type = "nombre";
        phase.label = "phase";
        phase.titre = "phase";
        phase.acquisition = "champ";
        phase.defaut = "0";
        addParamIn(phase);
        
        ParamIn bruit = new ParamIn();
        bruit.type = "nombre";
        bruit.label = "bruit";
        bruit.titre = "bruit";
        bruit.acquisition = "champ";
        bruit.defaut = "0";
        addParamIn(bruit);
        
       // paramtres de sortie
        ParamOut fx = new ParamOut();
        fx.type = "nombre";
        fx.label = "fx";
        fx.titre = "tps (s)";
        addParamOut(fx);
        
        ParamOut fy = new ParamOut();
        fy.type = "nombre";
        fy.label = "fy";
        fy.titre = "";
        addParamOut(fy);
        
        ParamOut specx = new ParamOut();
        specx.type = "nombre";
        specx.label = "specx";
        specx.titre = "frquence (Hz)";
        addParamOut(specx);
        
        ParamOut specy = new ParamOut();
        specy.type = "nombre";
        specy.label = "specy";
        specy.titre = "";
        addParamOut(specy);
        
        ParamOut info = new ParamOut();
        info.type = "string";
        info.label = "info";
        addParamOut(info);
        
        // options d'affichage
        Affichage multi = new Affichage();
        multi.type = "multiplot";
        multi.titre = "Transforme de Fourier";
        multi.sousAff = new Affichage[2];
        
        graphe1 = new Affichage();
        graphe1.type = "plot";
        graphe1.titre = "fonction";
        String[][] pparams1 = {{"fx","fy"}};
        graphe1.params = pparams1;
        graphe1.points = "none";
        graphe1.connecter = true;
        graphe1.effacer = true;
        graphe1.ajuster = true;
        multi.sousAff[0] = graphe1;
        
        graphe2 = new Affichage();
        graphe2.type = "plot";
        graphe2.titre = "spectre |TF|^2";
        String[][] pparams2 = {{"specx","specy"}};
        graphe2.params = pparams2;
        graphe2.points = "none";
        graphe2.connecter = true;
        graphe2.effacer = true;
        graphe2.ajuster = true;
        multi.sousAff[1] = graphe2;
        
        addAffichage(multi);
        super.init();
}


    public void initCalculLive(ListeValeurs in) throws SimuException {
        throw new SimuException("calculLive n'est pas implment dans TF: " +
            "utiliser calcul  la place");
    }
    
    public ListeValeurs calculLive() throws SimuException {
        throw new SimuException("calculLive n'est pas implment dans TF: " +
            "utiliser calcul  la place");
    }
    
    public ListeValeurs calcul(ListeValeurs in) throws SimuException {
        String fonction = in.lireString("fonction");
		double duree = in.lireDouble("duree");
        double periode = in.lireDouble("periode");
        double periode2 = in.lireDouble("periode2");
        String methode = in.lireString("methode");
        double amplitude = in.lireDouble("amplitude");
        double phase = in.lireDouble("phase");
        double bruit = in.lireDouble("bruit");
		if (periode <= 0)
			periode = 0.001;
        int npts = in.lireEntier("npts");
		if (npts < 1)
			npts = 1;
        
		double[] tfx = calcul_fx(npts, 0., duree);
		double[] tfy = calcul_fy(fonction, periode, periode2, tfx, amplitude, phase, bruit);
		double[] tspecx = calcul_specx(tfx, duree);
		double[] tspecy = calcul_specy(tfx, tfy, tspecx, methode);
		
        ListeValeurs out = new ListeValeurs();
        for (int i=0; i<npts; i++) {
            out.ajouterDouble("fx", tfx[i]);
            out.ajouterDouble("fy", tfy[i]);
		}
        for (int i=0; i<tspecx.length; i++) {
            out.ajouterDouble("specx", tspecx[i]);
            out.ajouterDouble("specy", tspecy[i]);
		}
		
        // sortie
        //graphe1.xdebut = "0";
        //graphe1.xfin = "" + dm;
        //graphe1.ydebut = "" + min(intrf);
        //graphe1.yfin = "" + max(intrf);
        
		String info = "rsolution: " + joliFormattage(1 / duree) + " Hz";
		double dt = duree / (npts - 1);
		double fc = 1 / (2*dt);
		info += "  frquence de coupure: " + joliFormattage(fc) + " Hz";
		out.ajouterString("info", info);
		
        return out;
    }
    
	// mthode d'affichage de double issue du Calcotron
    public static String joliFormattage(double v) {
        int precision = 3;
        int limitesci = 3; // dcimales avant de passer en mode sci
        
        if (v == 0)
            return("0");
        String sj = ""+v;
        if ("Infinity".equals(sj) || "-Infinity".equals(sj) || "NaN".equals(sj))
            return(""+v);
        String signe;
        if (v >= 0)
            signe = "";
        else {
            v = -v;
            signe = "-";
        }
        int exposant = (int)Math.floor(Math.log(v)/Math.log(10));
        double exparr = Math.pow(10, precision);
        double mantisse = v / Math.pow(10, exposant);
        mantisse = Math.rint(mantisse*exparr)/exparr;
        if (mantisse >= 10) {
            exposant += 1;
            mantisse = v / Math.pow(10, exposant);
            mantisse = Math.rint(mantisse*exparr)/exparr;
        }
        String sv = "";
        int chiffre = (int)Math.floor(mantisse);
        sv += chiffre + ".";
        for (int i=0; i<precision-1; i++) {
            mantisse = (mantisse - chiffre)*10;
            double exparr2 = Math.pow(10, precision-i);
            mantisse = Math.rint(mantisse*exparr)/exparr;
            chiffre = (int)Math.floor(mantisse);
            sv += chiffre;
        }
        mantisse = (mantisse - chiffre)*10;
        chiffre = (int)Math.round(mantisse);
        sv += chiffre;
        char dernier = sv.charAt(sv.length()-1);
        while (dernier == '0') {
            sv = sv.substring(0, sv.length()-1);
            dernier = sv.charAt(sv.length()-1);
        }
        if (exposant != 0) {
            if (exposant > 0 && exposant < limitesci) {
                String sd = sv.substring(0,1);
                if (sv.length()-2 < exposant) {
                    // transformation de 3.2E2 en 320
                    if (sv.length() > 2)
                        sd += sv.substring(2,sv.length());
                    for (int i=0; i<exposant-(sv.length()-2); i++)
                        sd += '0';
                } else {
                    // transformation de 3.245E2 en 324.5
                    if (sv.length() < exposant+2)
                        sd += sv.substring(2);
                    else
                        sd += sv.substring(2,exposant+2) + "." + sv.substring(exposant+2);
                }
                sv = sd;
            } else if (exposant < 0 && exposant > -limitesci) {
                // transformation de 3.2E-2 en 0.032
                String sd = "0.";
                for (int i=0; i<-exposant-1; i++)
                    sd += '0';
                sv = sd + sv.charAt(0) + sv.substring(2);
            } else
                sv += "E" + exposant;
        }
        if (sv.charAt(sv.length()-1) == '.')
            sv = sv.substring(0, sv.length()-1);
        
        sv = signe + sv;
        
        return(sv);
    }
    
	public double[] calcul_fx(int npts, double x0, double x1) {
		double[] tfx = new double[npts];
		for (int i=0; i<npts; i++)
			tfx[i] = x0 + i*(x1 - x0)/(npts - 1);
		return(tfx);
	}
	
	public double[] calcul_fy(String fonction, double periode, double periode2, double[] tfx, double amplitude,
			double phase, double bruit) {
		double[] tfy;
		
		if ("porte".equals(fonction))
			tfy = fonction_porte(tfx, amplitude, phase, bruit);
		else if ("crneau".equals(fonction))
			tfy = fonction_creneau(tfx, periode, amplitude, phase, bruit);
		else if ("sinusode".equals(fonction))
			tfy = fonction_sinusoide(tfx, periode, amplitude, phase, bruit);
		else if ("2 sinusodes".equals(fonction))
			tfy = fonction_double_sinusoide(tfx, periode, periode2, amplitude, phase, bruit);
		else if ("fichier".equals(fonction))
			tfy = fonction_fichier(tfx);
		else
			tfy = null;
		
		return(tfy);
	}
	
	public double[] calcul_specx(double[] tfx, double duree) {
		double[] tspecx = new double[tfx.length/2];
		//double[] tspecx = new double[tfx.length];
		double dn = 1/duree;
		for (int i=0; i<tspecx.length; i++)
			tspecx[i] = i*dn;
		return(tspecx);
	}
	
	public double[] calcul_specy(double[] tfx, double[] tfy, double[] tspecx, String methode) {
		double[] tspecy = new double[tspecx.length];
		if ("FFT".equals(methode)) {
			/*
			Complex[] fcomp = new Complex[tfx.length];
			for (int i=0; i<tfx.length; i++)
				fcomp[i] = new Complex(tfy[i], 0);
			Complex[] res = dif(fcomp);
			for (int i=0; i<tspecx.length; i++)
				tspecy[i] = res[i].mod2();
			*/
			// arrondi suprieur pour avoir une puissance de 2
			int n = (int)Math.pow(2, Math.ceil(Math.log(tfx.length)/Math.log(2)));
			if (n != tfx.length)
				System.out.println("FFT: " + tfx.length + " arrondi a " + n + ", remplissage avec des zeros");
			double[] f_r = new double[n];
			double[] f_i = new double[n];
			for (int i=0; i<n; i++) {
				if (i < tfy.length)
					f_r[i] = tfy[i];
				else
					f_r[i] = 0;
				f_i[i] = 0;
			}
			TFLib.fft_rec(false, f_r, f_i);
			// le padding pose des problmes...
			//if (n != tfx.length)
			//	for (int i=0; i<n/2; i+=2)
			//		tspecy[i/2] = f_r[i]*f_r[i] + f_i[i]*f_i[i];
			//else
				for (int i=0; i<tspecx.length; i++)
					tspecy[i] = f_r[i]*f_r[i] + f_i[i]*f_i[i];
		} else {
			double temp, temp_cos, temp_sin;
			for (int i=0; i<tspecx.length; i++) {
				//temp_cos = transfo_cos(tfx, tfy, tspecx[i]);
				//temp_sin = transfo_sin(tfx, tfy, tspecx[i]);
				temp_cos = 0;
				temp_sin = 0;
				for (int k=0; k<tfx.length; k++) {
					temp = 2 * Math.PI * tfx[k] * tspecx[i];
					temp_cos += tfy[k] * Math.cos(temp);
					temp_sin += tfy[k] * Math.sin(temp);
				}
	            tspecy[i] = temp_cos*temp_cos + temp_sin*temp_sin;
			}
		}
		return(tspecy);
	}
	
		
	/*
    public static double transfo_cos(double[] tfx, double[] tfy, double specx) {
		double somme = 0.;
		for (int i=0; i<tfx.length; i++)
			somme += tfy[i] * Math.cos(2 * Math.PI * tfx[i] * specx);
		return(somme);
	}
	
    public static double transfo_sin(double[] tfx, double[] tfy, double specx) {
		double somme = 0.;
		for (int i=0; i<tfx.length; i++)
			somme += tfy[i] * Math.sin(2 * Math.PI * tfx[i] * specx);
		return(somme);
	}
	*/
	
	public static double[] fonction_porte(double[] tfx, double amplitude, double phase, double bruit) {
		double[] tfy = new double[tfx.length];
		for (int i=0; i<tfx.length; i++) {
			if (tfx[i] < phase || tfx[i] >= phase + tfx[tfx.length - 1]/2)
				tfy[i] = 0.;
			else
				tfy[i] = amplitude;
			tfy[i] += bruit * Math.random();
		}
		return(tfy);
	}
	
	public static double[] fonction_creneau(double[] tfx, double periode, double amplitude, double phase, double bruit) {
		double[] tfy = new double[tfx.length];
		for (int i=0; i<tfx.length; i++) {
			double xmod = (tfx[i] + phase) / periode;
			xmod = (xmod - Math.floor(xmod)) * periode;
			if (xmod < periode/2)
				tfy[i] = 0;
			else
				tfy[i] = amplitude;
			tfy[i] += bruit * Math.random();
		}
		return(tfy);
	}
	
	public static double[] fonction_sinusoide(double[] tfx, double periode, double amplitude, double phase, double bruit) {
		double[] tfy = new double[tfx.length];
		for (int i=0; i<tfx.length; i++)
			tfy[i] = amplitude*Math.sin( (tfx[i] + phase) * 2 * Math.PI / periode ) + bruit * Math.random();
		return(tfy);
	}
	
	public static double[] fonction_double_sinusoide(double[] tfx, double periode, double periode2, double amplitude,
			double phase, double bruit) {
		double[] tfy = new double[tfx.length];
		for (int i=0; i<tfx.length; i++)
			tfy[i] = Math.sin( tfx[i] * 2 * Math.PI / periode ) +
				amplitude*Math.sin( (tfx[i] + phase) * 2 * Math.PI / periode2 ) +
					bruit * Math.random();
		return(tfy);
	}
	
	public double[] fonction_fichier(double[] tfx) {
		/*
        // lecture du fichier
        String cheminf;
        if (dossier != null)
            cheminf = dossier + File.separator + fspectre;
        else
            cheminf = fspectre;
        double[][] intensites = lireFichier(cheminf);
        if (intensites == null)
            return(null);
        double[] s = intensites[0];
        double[] fs = intensites[1];
		*/
		return(null);
	}
	
    class EnteteFichier {
        public String titre;
        public double bornemin, bornemax;
    }
    
    /**
     * lire l'entte d'un fichier ASCII venant d'IDL,
     * avec le nombre d'lments sur la premire ligne
     * les bornes  et l'incrment inc sur la deuxime ligne
     * la position gnrale du spectre (ignore) sur la troisime ligne
     * et "y" normalis sur les suivantes
     */
    public EnteteFichier lireEnteteFichier(String nomFichier) {
        try {
            URL url = new URL(getDocumentBase(), nomFichier);
            BufferedReader in = new BufferedReader(new InputStreamReader(url.openStream()));
            String ligne = in.readLine();
            ligne = in.readLine();
            ligne = ligne.trim();
            int is = ligne.indexOf(' ');
            String sx = ligne.substring(0,is).trim();
            double bornemin = (Double.valueOf(sx)).doubleValue();
            ligne = ligne.substring(is+1).trim();
            is = ligne.indexOf(' ');
            String sy = ligne.substring(0,is).trim();
            double bornemax = (Double.valueOf(sy)).doubleValue();
            ligne = ligne.substring(is+1).trim();
            double increment = (Double.valueOf(ligne)).doubleValue();
            ligne = in.readLine();
            String titrepos = ligne.trim();
            in.close();
            EnteteFichier ent = new EnteteFichier();
            ent.titre = titrepos;
            ent.bornemin = bornemin;
            ent.bornemax = bornemax;
            return(ent);
        } catch (MalformedURLException ex) {
            System.err.println(ex.getClass().getName() + " : " + ex.getMessage());
            return(null);
        } catch (NumberFormatException ex) {
            System.err.println(ex.getClass().getName() + " : " + ex.getMessage());
            return(null);
        } catch (IOException ex) {
            System.err.println(ex.getClass().getName() + " : " + ex.getMessage());
            return(null);
        }
    }
    
    /**
     * lire les donnes d'un fichier ASCII venant d'IDL,
     * avec le nombre d'lments sur la premire ligne
     * les bornes  et l'incrment inc sur la deuxime ligne
     * la position gnrale du spectre (ignore) sur la troisime ligne
     * et "y" normalis sur les suivantes
     * -> t[0][i] = bornemin + increment*i;
     *    t[1][i] = y;
     *    t[0][t[0].length-1] = bornemax; (bornemin + inc*n risque d'tre trop grand)
     */
    public double[][] lireFichier(String nomFichier) {
        try {
            URL url = new URL(getDocumentBase(), nomFichier);
            BufferedReader in = new BufferedReader(new InputStreamReader(url.openStream()));
            String ligne = in.readLine();
            int n = Integer.parseInt(ligne.trim());
            ligne = in.readLine();
            ligne = ligne.trim();
            int is = ligne.indexOf(' ');
            String sx = ligne.substring(0,is).trim();
            double bornemin = (Double.valueOf(sx)).doubleValue();
            ligne = ligne.substring(is+1).trim();
            is = ligne.indexOf(' ');
            String sy = ligne.substring(0,is).trim();
            double bornemax = (Double.valueOf(sy)).doubleValue();
            ligne = ligne.substring(is+1).trim();
            double increment = (Double.valueOf(ligne)).doubleValue();
            ligne = in.readLine();
            double[][] t = new double[2][n];
            double y;
            for (int i=0; i<n && ligne != null; i++) {
                ligne = in.readLine();
                if (ligne != null && !"".equals(ligne)) {
                    ligne = ligne.trim();
                    y = (Double.valueOf(ligne)).doubleValue();
                    t[0][i] = bornemin + increment*i;
                    t[1][i] = y;
                }
            }
            in.close();
            // pour viter les erreurs d'arrondi
            t[0][t[0].length-1] = bornemax;
            return(t);
        } catch (MalformedURLException ex) {
            System.err.println(ex.getClass().getName() + " : " + ex.getMessage());
            return(null);
        } catch (NumberFormatException ex) {
            System.err.println(ex.getClass().getName() + " : " + ex.getMessage());
            return(null);
        } catch (IOException ex) {
            System.err.println(ex.getClass().getName() + " : " + ex.getMessage());
            return(null);
        }
    }
    
    public static double min(double[] t) {
        double min = t[0];
        for (int i=1; i<t.length; i++)
            if (t[i] < min)
                min = t[i];
        return(min);
    }
    
    public static double max(double[] t) {
        double max = t[0];
        for (int i=1; i<t.length; i++)
            if (t[i] > max)
                max = t[i];
        return(max);
    }
    
}
