/*
Bibliothque de fonctions pour calculer des Transformes de Fourier

Copyright (C) 2004 Observatoire de Paris-Meudon

Ce programme est un logiciel libre ; vous pouvez le redistribuer et/ou le modifier conformment aux dispositions de la Licence Publique Gnrale GNU, telle que publie par la Free Software Foundation ; version 2 de la licence, ou encore ( votre choix) toute version ultrieure.

Ce programme est distribu dans l'espoir qu'il sera utile, mais SANS AUCUNE GARANTIE ; sans mme la garantie implicite de COMMERCIALISATION ou D'ADAPTATION A UN OBJET PARTICULIER. Pour plus de dtail, voir la Licence Publique Gnrale GNU.

Vous devez avoir reu un exemplaire de la Licence Publique Gnrale GNU en mme temps que ce programme ; si ce n'est pas le cas, crivez  la Free Software Foundation Inc., 675 Mass Ave, Cambridge, MA 02139, Etats-Unis.
*/

//import Complex;

/**
 * Bibliothque de fonctions pour calculer des Transformes de Fourier
 * compilation: javac -target 1.1 -encoding ISO-8859-1 TFLib.java
 */
public class TFLib {
    
	protected double[][] tw_cos;
	protected double[][] tw_sin;
	protected double[] f_r;
	protected double[] f_i;
	
	protected TFLib(double[] f_r, double[] f_i) {
		this.f_r = f_r;
		this.f_i = f_i;
	}
	
	/**
	 * FFT en 2D
	 */
    public static void transfo2D_FFT(double[][] imagep, double[][] figured, boolean log, boolean quadrants) {
		int dimx = imagep[0].length;
		int dimy = imagep.length;
		for (int i=0; i<dimx; i++)
			for (int j=0; j<dimy; j++)
				figured[i][j] = 0;
		double[][][] fk = new double[dimx][dimy][2];
		//long t1 = System.currentTimeMillis();
		for (int kx=0; kx<dimx; kx++) {
			double[] f_r = new double[dimy];
			double[] f_i = new double[dimy];
			for (int i=0; i<dimy; i++) {
				f_r[i] = imagep[kx][i];
				f_i[i] = 0;
			}
			
			fft_rec(false, f_r, f_i);
			//autre_fft(1, f_r, f_i);
			
			for (int i=0; i<dimy; i++) {
				fk[kx][i][0] = f_r[i];
				fk[kx][i][1] = f_i[i];
				//figured[kx][i] = Math.sqrt(f_r[i]*f_r[i] + f_i[i]*f_i[i]);
			}
		}
		for (int ky=0; ky<dimy; ky++) {
			double[] f_r = new double[dimx];
			double[] f_i = new double[dimx];
			for (int i=0; i<dimx; i++) {
				f_r[i] = fk[i][ky][0];
				f_i[i] = fk[i][ky][1];
			}
			
			fft_rec(false, f_r, f_i);
			//autre_fft(1, f_r, f_i);
			
			double v;
			for (int i=0; i<dimx; i++) {
				if (log)
					v = Math.log(Math.sqrt(f_r[i]*f_r[i] + f_i[i]*f_i[i]));
				else
					v = Math.sqrt(f_r[i]*f_r[i] + f_i[i]*f_i[i]);
				if (quadrants) {
					// rorganisation des coordonnes
					int i2;
					if (i >= dimx/2)
						i2 = i - dimx/2;
					else
						i2 = i + dimx/2;
					int ky2;
					if (ky >= dimy/2)
						ky2 = ky - dimy/2;
					else
						ky2 = ky + dimy/2;
					figured[i2][ky2] = v;
				} else
					figured[i][ky] = v;
			}
		}
		//long t2 = System.currentTimeMillis();
		//System.out.println((t2-t1)/1000.0+" s");
	}
	
	/*
	// algo FFT le plus proche des quations, mais pas du tout optimis
	public static Complex[] dif(Complex[] f) {
		int n = f.length;
		if (n == 1) {
			return(f);
		}
		int np = n / 2;
		Complex[] fe = new Complex[np];
		Complex[] fo = new Complex[np];
		for (int i=0; i<np; i++) {
			// fe[i] = f[i] + f[i+np];
			fe[i] = Complex.addition(f[i], f[i+np]);
			double tmp = 2*Math.PI*((double)i)/n;
			Complex twiddle = new Complex(Math.cos(tmp), -Math.sin(tmp));
			//fo[i] = (f[i] - f[i+np]) * twiddle;
			fo[i] = Complex.multiplication(Complex.soustraction(f[i], f[i+np]), twiddle);
		}
		fe = dif(fe);
		fo = dif(fo);
		Complex[] res = new Complex[n];
		for (int i=0; i<np; i++) {
			res[2*i] = fe[i];
			res[2*i+1] = fo[i];
		}
		return(res);
	}
	*/
	
	protected void bit_reverse() {
		bit_reverse(f_r);
		bit_reverse(f_i);
	}
	
	protected static void bit_reverse(double[] t) {
		int j = 0;
		double temp;
		int m;
		//for (int i=0; i<t.length; i++) {
		for (int i=0; i<t.length-1; i++) {
			if (j > i) {
				temp = t[j];
				t[j] = t[i];
				t[i] = temp;
			}
			m = t.length / 2;
			//while (m >= 1 && j >= m) {
			while (j >= m) {
				j -= m;
				m /= 2;
			}
			j += m;
		}
	}
	
	/**
	 * FFT rcursive un peu optimise.
	 * la taille de f (f_r et f_i) doit tre une puissance de 2.
	 */
	 /*
	public static void fft_rec(boolean inverse, double[] f_r, double[] f_i) {
		dif(inverse, 0, f_r.length, f_r, f_i);
		bit_reverse(f_r);
		bit_reverse(f_i);
	}
	*/
	
	/**
	 * FFT rcursive un peu optimise.
	 * la taille de f (f_r et f_i) doit tre une puissance de 2.
	 * avec precalcul_tw.
	 */
	public static void fft_rec(boolean inverse, double[] f_r, double[] f_i) {
		TFLib lib = new TFLib(f_r, f_i);
		lib.precalcul_tw();
		int log2n = (int)(Math.log(f_r.length)/Math.log(2));
		lib.dif(inverse, 0, log2n);
		lib.bit_reverse();
	}
	
	protected void precalcul_tw() {
		int n = f_r.length;
		int log2n = (int)(Math.log(n)/Math.log(2));
		tw_cos = new double[n/2][log2n+1];
		tw_sin = new double[n/2][log2n+1];
		double tw;
		int m = n;
		int log2m = log2n;
		while (m > 1) {
			for (int i=0; i<m/2; i++) {
				tw = 2*Math.PI*((double)i)/m;
				tw_cos[i][log2m] = Math.cos(tw);
				tw_sin[i][log2m] = Math.sin(tw);
			}
			m /= 2;
			log2m--;
		}
	}
	
	/*
	protected static void dif(boolean inverse, int ind, int n, double[] f_r, double[] f_i) {
		int np = n >> 1;
		int ind2 = ind + np;
		double e_r, e_i, o_r, o_i;
		double tw, tw_r, tw_i, tmp_r, tmp_i;
		for (int i=0; i<np; i++) {
			// e = f[ind+i] + f[ind2+i];
			e_r = f_r[ind+i] + f_r[ind2+i];
			e_i = f_i[ind+i] + f_i[ind2+i];
			//double tw = 2*Math.PI*tspecx[i];
			tw = 2*Math.PI*((double)i)/n;
			//tw_r = tw_cos[i][log2n];
			//tw_i = -tw_sin[i][log2n];
			tw_r = Math.cos(tw);
			tw_i = -Math.sin(tw);
			if (inverse)
				tw_i = -tw_i;
			tmp_r = f_r[ind+i] - f_r[ind2+i];
			tmp_i = f_i[ind+i] - f_i[ind2+i];
			// o = (f[ind+i] - f[ind2+i]) * e^(-2*j*pi*i/n);
			o_r = tmp_r * tw_r - tmp_i * tw_i;
			o_i = tmp_r * tw_i + tmp_i * tw_r;
			// f[ind+i] = e;
			f_r[ind+i] = e_r;
			f_i[ind+i] = e_i;
			// f[ind2+i] = o;
			f_r[ind2+i] = o_r;
			f_i[ind2+i] = o_i;
		}
		if (np > 1) {
			dif(inverse, ind, np, f_r, f_i);
			dif(inverse, ind2, np, f_r, f_i);
		}
	}
	*/
	
	// avec precalcul_tw
	protected void dif(boolean inverse, int ind, int log2n) {
		int np = 1<< (log2n-1);
		int ind2 = ind + np;
		double e_r, e_i, o_r, o_i;
		double tw_r, tw_i, tmp_r, tmp_i;
		
		for (int i=0; i<np; i++) {
			// e = f[ind+i] + f[ind2+i];
			e_r = f_r[ind+i] + f_r[ind2+i];
			e_i = f_i[ind+i] + f_i[ind2+i];
			
			//double tw = 2*Math.PI*tspecx[i];
			//tw = 2*Math.PI*((double)i)/n;
			//tw_r = Math.cos(tw);
			//tw_i = -Math.sin(tw);
			tw_r = tw_cos[i][log2n];
			tw_i = -tw_sin[i][log2n];
			if (inverse)
				tw_i = -tw_i;
			
			tmp_r = f_r[ind+i] - f_r[ind2+i];
			tmp_i = f_i[ind+i] - f_i[ind2+i];
			
			// o = (f[ind+i] - f[ind2+i]) * e^(-2*j*pi*i/n);
			o_r = tmp_r * tw_r - tmp_i * tw_i;
			o_i = tmp_r * tw_i + tmp_i * tw_r;
			
			// f[ind+i] = e;
			f_r[ind+i] = e_r;
			f_i[ind+i] = e_i;
			
			// f[ind2+i] = o;
			f_r[ind2+i] = o_r;
			f_i[ind2+i] = o_i;
		}
		
		if (log2n > 1) {
			dif(inverse, ind, log2n-1);
			dif(inverse, ind2, log2n-1);
		}
	}
	
	/**
	 * FFT non rcursive
	 *
	 * avec astuces pour les "twiddle factors"
	 * T(N,n) = e^(-i*2*pi*n/N)
	 * T(N,0) = 0    T(N,N/4) = -i    T(N,N/8) = (1 - i)/sqrt(2)   T(N,3*N/8) = -(1 + i)/sqrt(2)
	 * prcision ?
	 */
	public static void autre_fft(int dir, double[] f_r, double[] f_i) {
		int nn,i,i1,j,k,i2,l,l1,l2;
		double c1,c2,t1,t2,u1,u2,z;
		nn = f_r.length;
		double m = Math.log(nn)/Math.log(2);
		
		bit_reverse(f_r);
		bit_reverse(f_i);
		
		c1 = -1.0;
		c2 = 0.0;
		l2 = 1;
		for (l=0; l<m; l++) {
		   l1 = l2;
		   l2 <<= 1;
		   u1 = 1.0;
		   u2 = 0.0;
		   for (j=0; j<l1; j++) {
		      for (i=j; i<nn; i+=l2) {
		         i1 = i + l1;
		         t1 = u1 * f_r[i1] - u2 * f_i[i1];
		         t2 = u1 * f_i[i1] + u2 * f_r[i1];
		         f_r[i1] = f_r[i] - t1;
		         f_i[i1] = f_i[i] - t2;
		         f_r[i] += t1;
		         f_i[i] += t2;
		      }
		      z =  u1 * c1 - u2 * c2;
		      u2 = u1 * c2 + u2 * c1;
		      u1 = z;
		   }
		   c2 = Math.sqrt((1.0 - c1) / 2.0);
		   if (dir == 1)
		      c2 = -c2;
		   c1 = Math.sqrt((1.0 + c1) / 2.0);
		}
	
		// Scaling for forward transform
		if (dir == 1) {
		   for (i=0;i<nn;i++) {
		      f_r[i] /= (double)nn;
		      f_i[i] /= (double)nn;
		   }
		}
	}
	
}
